import os
import requests

from ybc_config import config
from ybc_exception import InternalError

_COURSE_API_PREFIX = config["course-api-prefix"]
# 每个 HTTP 请求最多的请求次数
_MAX_RETRY_TIMES = 3


def _read_env(key):
    return os.environ[key] if key in os.environ else ''


def _build_headers():
    persistent = _read_env('REFRESH_TOKEN')
    headers = {
        'Cookie': 'persistent=' + persistent,
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    return headers


def _to_json(response):
    """
    从 HTTP 请求的响应中获取数据并解析成 JSON 格式
    异步方法: JS requests 库中基于 fetch api, 其 Response.json() 方法返回 Promise

    :param response:
    :return:
    """
    return response.json()


def get(url_rel_path, params=None):
    """
    带失败重试的 HTTP GET 方法

    :param url_rel_path: url 相对路径, 例如 animal-sound
    :param params: 请求参数
    :return: 请求返回的数据, JSON 对象
    """
    url = '{}/{}'.format(_COURSE_API_PREFIX, url_rel_path)
    headers = _build_headers()
    for i in range(_MAX_RETRY_TIMES):
        # TODO: 增加 headers, 注入 cookie 等
        response = requests.get(
            url=url,
            params=params if params else {},
            headers=headers
        )
        # TODO: 根据 status code 决定是否进行重试, 比如 4XX 不应该重试, 5XX 应该重试
        if response.ok:
            return _to_json(response)

    raise InternalError('调用 get 接口失败: url={}, params={}'.format(url, params), __name__)


def post(url_rel_path, data, params=None):
    """
    带失败重试的 HTTP POST 方法

    :param url_rel_path: url 相对路径, 例如 animal-sound
    :param data: 请求的 Body
    :param params: 请求参数
    :return: 请求返回的数据, JSON 对象
    """
    url = '{}/{}'.format(_COURSE_API_PREFIX, url_rel_path)
    headers = _build_headers()
    for i in range(_MAX_RETRY_TIMES):
        response = requests.post(
            url=url,
            json=data,
            params=params if params else {},
            headers=headers,
        )

        if response.ok:
            return _to_json(response)

    raise InternalError('调用 post 接口失败: url={}, data={}, params={}'.format(url, data, params),
                        __name__)



import sys
from ybc_exception import ParameterTypeError
from ybc_exception import ParameterValueError


class Argument:
    def __init__(self, module_name, method_name, name, value, type, predicates):
        self.module_name = module_name
        self.method_name = method_name
        self.name = name
        self.value = value
        self.type = type
        self.predicates = predicates

    def check_argument(self):
        self._check_type()
        self._check_value()

    def _check_type(self):
        if self.type is not None and not isinstance(self.value, self.type):
            raise ParameterTypeError(self.method_name, self.name)

    def _check_value(self):
        if self.predicates is not None and not self.predicates(self.value):
            raise ParameterValueError(self.method_name, self.name)


class Checker:
    @staticmethod
    def check_arguments(arguments):
        if sys.platform != 'skulpt':
            return
        for argument in arguments:
            argument.check_argument()



import os
import sys
import requests

from ybc_config import config

if sys.platform == 'win32':
    HOME_DRIVE_KEY = 'HOMEDRIVE'
    HOME_PATH_KEY = 'HOMEPATH'
    _DEFAULT_USER_WORKSPACE_ROOT = (
        r'{}{}\AppData\Roaming\Prometheus\Application Data\workspace'
        .format(os.environ[HOME_DRIVE_KEY], os.environ[HOME_PATH_KEY])
    )
else:
    _DEFAULT_USER_WORKSPACE_ROOT = '/sandbox'
_USER_WORKSPACE_ROOT = os.environ.get('USER_WORKSPACE_ROOT', _DEFAULT_USER_WORKSPACE_ROOT)
_OSS_ENDPOINT = config['prefix-oss']
_PUBLIC_FILES_UPLOAD_API = config['course-api-prefix'] + "/public-files"


def url_of(filename, is_new_file=False):
    """
    获取文件在 OSS 上的 URL

    :param filename: 文件相对于当前工作目录的路径, 比如 '1.jpg', '../1.jpg', './1.jpg' 等
    :param is_new_file: 是否是待创建的新文件, 如果是待创建的新文件, 则不需要检查文件是否存在
    :return: 文件在 OSS 上的 URL
    """
    if not filename or (filename.startswith('https://')
                        or filename.startswith('http://')):
        return filename
    elif os.path.isdir(filename):
        # 不对文件夹进行转换，只判断是不是存在的文件夹
        return filename
    elif is_new_file:
        # 获取文件在文件系统中的绝对路径
        local_file_path = os.path.abspath(filename)
        # 获取文件相对于用户根目录的相对路径
        # 比如: 41314233/default/turtle/wc/1.jpg
        oss_file_path = os.path.relpath(local_file_path, _USER_WORKSPACE_ROOT).replace('\\', '/')
        # 拼接出文件在 OSS 上的 URL
        return '{}/{}'.format(_OSS_ENDPOINT, oss_file_path)
    else:
        with open(filename, 'rb') as f:
            # 上传文件，生成 URL
            response = requests.post(_PUBLIC_FILES_UPLOAD_API, files={'file': f})
            if response.ok:
                public_file = response.json()
                return public_file['url']
            else:
                raise RuntimeError()



# only for skulpt
class ArgumentsChecker:
    def __init__(self, predicates):
        self.predicates = predicates

    def __call__(self, fn):
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)
        return wrapper



import sys
if sys.platform == 'skulpt':
    class Predicate(object):
        def __init__(self, fn):
            self.fn = fn

        def __call__(self, *args, **kwargs):
            return self.fn(*args, **kwargs)

        def __and__(self, other):
            return AndPredicate(self, other)

        def __rand__(self, other):
            return AndPredicate(other, self)

        def __or__(self, other):
            return OrPredicate(self, other)

        def __ror__(self, other):
            return OrPredicate(other, self)

        def __invert__(self):
            return InvertedPredicate(self)

        def after(self, transform):
            return PostPredicate(self, transform)
else:
    from functools import update_wrapper
    from typing import Callable

    class Predicate(Callable):
        def __init__(self, fn):
            self.fn = fn
            update_wrapper(self, fn)

        def __call__(self, *args, **kwargs):
            return self.fn(*args, **kwargs)

        def __and__(self, other):
            return AndPredicate(self, other)

        def __rand__(self, other):
            return AndPredicate(other, self)

        def __or__(self, other):
            return OrPredicate(self, other)

        def __ror__(self, other):
            return OrPredicate(other, self)

        def __invert__(self):
            return InvertedPredicate(self)

        def after(self, transform):
            return PostPredicate(self, transform)


class ComposedPredicate(Predicate):
    def __init__(self, fn, left, right):
        super().__init__(fn)
        self.left = left
        self.right = right


class TransformedPredicate(Predicate):
    def __init__(self, fn, prev):
        super().__init__(fn)
        self.prev = prev


class AndPredicate(ComposedPredicate):
    def __init__(self, left, right):
        def composed(*args, **kwargs):
            return left(*args, **kwargs) and right(*args, **kwargs)

        super().__init__(composed, left, right)


class OrPredicate(ComposedPredicate):
    def __init__(self, left, right):
        def composed(*args, **kwargs):
            return left(*args, **kwargs) or right(*args, **kwargs)

        super().__init__(composed, left, right)


class InvertedPredicate(TransformedPredicate):
    def __init__(self, prev):
        def transformed(*args, **kwargs):
            return not prev(*args, **kwargs)

        super().__init__(transformed, prev)


class PostPredicate(TransformedPredicate):
    def __init__(self, prev, transform):
        def transformed(*args, **kwargs):
            return prev(transform(*args, **kwargs))

        super().__init__(transformed, prev)



import inspect
from functools import wraps
from inspect import Parameter
from typing import Callable

from ybc_exception import ParameterTypeError
from ybc_exception import ParameterValueError


def check_argument_types(fn, parameters, arguments):
    illegal_arguments = [
        (name, value)
        for name, value in arguments
        if (name in parameters
            and parameters[name].annotation is not Parameter.empty
            and not isinstance(value, parameters[name].annotation))
    ]

    if illegal_arguments:
        raise ParameterTypeError(
            fn.__name__,
            '、'.join("'%s'" % k for k, v in illegal_arguments))


def check_argument_values(fn, predicates, arguments):
    illegal_arguments = [
        (name, value)
        for name, value in arguments
        if name in predicates and not predicates[name](value)
    ]

    if illegal_arguments:
        raise ParameterValueError(
            fn.__name__,
            '、'.join("'%s'" % k for k, v in illegal_arguments))


class ArgumentsChecker:
    def __init__(self, predicates):
        self.predicates = predicates

    def __call__(self, fn: Callable):
        sig = inspect.signature(fn)

        @wraps(fn)
        def wrapper(*args, **kwargs):
            arguments = sig.bind(*args, **kwargs).arguments.items()

            check_argument_types(fn, sig.parameters, arguments)
            check_argument_values(fn, self.predicates, arguments)

            return fn(*args, **kwargs)

        return wrapper



import os
import re

from ybc_commons.oss import path
from ybc_exception import FilenameTooLongError
from ybc_exception import IllegalFilenameError
from . import Predicate


@Predicate
def true(*_):
    return True


@Predicate
def false(*_):
    return False


@Predicate
def is_truthy(value):
    return bool(value)


@Predicate
def is_falsy(value):
    return not value


@Predicate
def is_empty(value):
    return not value


@Predicate
def non_empty(value):
    return bool(value)


@Predicate
def non_blank(value):
    return bool(str.strip(value))


@Predicate
def is_valid_filename(filename):
    # 合法 url 跳过检查
    if re.match(r'(http[s]?://([\w-]+\.)+(([\w\-/.?%&=])*))', filename):
        return True

    # _____temp 文件夹跳过路径检查
    # 传入路径时先判断文件夹是否存在，不存在抛出 FileNotFoundError，存在则 split 逐项检测是否合法
    if '_____temp' not in filename:
        path.url_of(os.path.dirname(filename), is_new_file=False)

    names = filename.split("/")
    for name in names:
        if not re.match(r'^[\u4e00-\u9fa5\w.\-()]+$', name):
            raise IllegalFilenameError("'" + filename + "'")

        # 文件名长度限制不包含后缀名
        name = os.path.splitext(name)[0]
        if not len(re.sub(r'[\u4e00-\u9fa5]', 'aa', name)) < 41:
            raise FilenameTooLongError(("'" + filename + "'"))
    return True


@Predicate
def is_function(value):
    return callable(value)


def match_regex_ignorecase(regex_str):
    @Predicate
    def inner(value):
        return bool(re.match(regex_str, value, re.IGNORECASE))

    return inner


def is_self(self):
    @Predicate
    def inner(value):
        return value is self

    return inner


def equals(other):
    @Predicate
    def inner(value):
        return value == other

    return inner


def is_instance_of(type):
    @Predicate
    def inner(value):
        return isinstance(value, type)

    return inner


def is_greater_than(lower):
    @Predicate
    def inner(value):
        return value > lower

    return inner


def is_greater_or_equal(lower):
    @Predicate
    def inner(value):
        return value >= lower

    return inner


def is_less_than(upper):
    @Predicate
    def inner(value):
        return value < upper

    return inner


def is_less_or_equal(upper):
    @Predicate
    def inner(value):
        return value <= upper

    return inner


def is_in(iterable):
    @Predicate
    def inner(value):
        return value in iterable

    return inner


def is_in_range(lower, upper):
    @Predicate
    def inner(value):
        return lower <= value < upper

    return inner


def has_len_in_range(lower, upper):
    @Predicate
    def inner(value):
        return lower <= len(value) < upper

    return inner


def file_format_in(*formats):
    @Predicate
    def inner(filename):
        return any(filename.lower().endswith('.' + format) for format in formats)

    return inner


@Predicate
def is_image_file(filename):
    image_file_extensions = ['.jpg', '.jpeg', '.png']
    return any(filename.lower().endswith(extension) for extension in image_file_extensions)


@Predicate
def is_audio_file(filename):
    audio_file_extensions = ['.mp3', '.wav', '.mp4']
    return any(filename.lower().endswith(extension) for extension in audio_file_extensions)



import sys

if sys.platform == 'skulpt':
    from .argument_check_for_skulpt import ArgumentsChecker

    def check_arguments(fn_or_predicates):
        if callable(fn_or_predicates):
            fn = fn_or_predicates
            return ArgumentsChecker({})(fn)
        else:
            predicates = fn_or_predicates
            return ArgumentsChecker(predicates)

else:
    from .argument_check import ArgumentsChecker
    from functools import wraps
    from typing import Callable

    def check_arguments(fn_or_predicates):
        if isinstance(fn_or_predicates, Callable):
            fn = fn_or_predicates
            return ArgumentsChecker({})(fn)
        else:
            predicates = fn_or_predicates
            return ArgumentsChecker(predicates)

    class ReturnValue(Exception):
        def __init__(self, value):
            self.value = value

    def catch_return_value(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except ReturnValue as rv:
                return rv.value

        return wrapper
